<?php
/** \file mo_auto_add_terms.admin.inc
 *
 * \brief Administrative functions of the MO Auto add terms module.
 *
 * Implementation of the administrative functions used by
 * the Drupal auto-add-terms module.
 *
 * \section copyright Copyright (c) 2008-2012 Made to Order Software Corp.
 *
 * http://www.m2osw.com/mo_autoaddterms
 * contact@m2osw.com
 */

/**
 * \brief Number of nodes to process by CRON.
 *
 * Default number of nodes to process whenever some terms were added.
 *
 * Adding a term to any taxonomy that was set up with auto-add-terms
 * has the side effect of deleting the cached nodes.
 *
 * The result is that the CRON task will reprocess all the nodes and
 * automatically add the new terms to all the existing nodes, over
 * time.
 */
define('AUTOADDTERMS_NODE_COUNT', 10);


/**
 * \brief Add all the terms found in a node.
 *
 * This function is the one used to process one node.
 *
 * It checks whether the node was modified. If not, and the terms
 * where already cached, then nothing happens. Otherwise, it
 * attempts to find all the terms that are defined in the vocabularies
 * of this node and automatically attaches them to the node.
 *
 * For instance, if you have a sentence such as "I love Linux" and
 * have a vocabulary with the word "Linux" in it, whether or not you
 * put it in the list of tags for that node, it will automatically
 * be added by this function.
 *
 * Note that the function is case insensitive. Thus you can write
 * Linux in one and linux in the other and it will still work.
 *
 * This function is automatically called whenever you save a node
 * and by the cron task.
 *
 * \bug
 * Although probably rare, this function does not properly manage
 * the case when two different users/processes (i.e. a user and
 * CRON) work on the exact same node. This means the function may
 * fail once in a while, yet very rarely.
 *
 * \param[in,out] $node The node to be checked, new term list is returned
 * \param[in] $reset_cache Whether the cache of this node needs to be reset
 * \param[in] $is_cron Whether the function is being called from hook_cron()
 */
function _mo_auto_add_terms_process_node(&$node, $reset_cache = FALSE, $is_cron = FALSE) {
  if (!$reset_cache) {
    $result = db_query("SELECT nid FROM {mo_auto_add_terms_nodes} WHERE nid = %d", $node->nid);
    if ($row = db_fetch_array($result)) {
      // this node was cached already
      return;
    }
  }

  // total number of nodes processed (the same ones will be reprocessed over and over again...)
  variable_set('mo_auto_add_terms_count', variable_get('mo_auto_add_terms_count', 0) + 1);

  // we insert the node to the cache; that way, if it fails, we don't try
  // again and again for nothing; we use a transaction for the hell of it 8-)
  db_query("BEGIN");
  // the DELETE is required if the $reset_cache is TRUE
  db_query("DELETE FROM {mo_auto_add_terms_nodes} WHERE nid = %d", $node->nid);
  db_query("INSERT INTO {mo_auto_add_terms_nodes} (nid) VALUES (%d)", $node->nid);
  db_query("COMMIT");

  // Get the list of vocabularies attached to this node
  $sql = "SELECT vid FROM {vocabulary_node_types} WHERE type = '%s'";
  $results = db_query($sql, $node->type);
  while ($row = db_fetch_array($results)) {
    $vid = $row['vid'];
    $auto_add_terms_per_contenttype = variable_get('mo_auto_add_terms_auto_add_terms_per_contenttype' . $vid, array());
    if (_mo_auto_add_terms_selected($vid, 'test')
    && (!isset($auto_add_terms_per_contenttype[$node->type]) || $auto_add_terms_per_contenttype[$node->type])
    && (!$is_cron || _mo_auto_add_terms_autocheck($vid, 'test'))) {
      // the user authorized auto-add-terms on this vocabulary
      // and this node is attached to it!
      $terms = _mo_auto_add_terms_getallterms($vid);
      _mo_auto_add_terms_to_node($vid, $node, $terms);
    }
  }
}

/**
 * \brief Clean up a string of special characters
 *
 * This function is used to clean up a string from special
 * characters. It is used to clean up terms and the
 * node body, title, comments.
 *
 * The function also ensures that no double spaces are
 * found anywhere. All words will be separated by at most
 * one space. In that regard, it knows how to handle
 * &nbsp; properly.
 *
 * Replace any character other than a letter into a space
 * (we keep dashes and apostrophes though, so "I'm" and
 * "Auto-add" will match, which is a problem in other
 * languages like "l'école" in French which won't match
 * "école"... this can be solved with proper synonyms,
 * but that's tedious work!)
 *
 * The function then returns the string with one space
 * before, one space after and the string trimmed in
 * between.
 *
 * \note
 * The string is transformed to lowercase so the matching
 * does not need to be case insensitive.
 *
 * \todo
 * Remove Unicode punctuation.
 *
 * \param[in] $string The string to be transformed
 * \param[in] $vid The vocabulary identifier being checked
 *
 * \return The cleaned up string
 */
function mo_autoaddterm_clean_term($string, $vid) {
  static $user_pattern = array();
  if (!isset($user_pattern[$vid])) {
    $user_pattern[$vid] = str_replace('%', '\\%', variable_get('mo_auto_add_terms_ignore_pattern' . $vid, '[ -&(-,./:-?[-_{-~]'));
  }
  // XXX -- we need better handling of entities, i.e. &eacute; should match é
  return ' ' . trim(preg_replace(
    array(
      '%--+%',  // -- or --- <=> &mdash;
      '%(?:[\t\n\r]|&mdash;|&amp;)+%', // new line, &mdash;, %amp; transformed to one separator each
      '% -- (?:-- )+%', // -- -- => --
      '%&nbsp;%', // &nbsp; <=> ' '
      '%' . $user_pattern[$vid] . '%',
      '%  +%'
    ),
    array(
      ' -- ',
      ' -- ',
      ' -- ',
      ' ',
      ' ',
      ' ',
    ),
    drupal_strtolower($string)
  )) . ' ';
}

/**
 * \brief Match the words in the node with the words in the taxonomy.
 *
 * Search for all the terms in the node title and body
 * and add them to the vocabulary of that node.
 *
 * By default, terms are removed unless present in the node.
 *
 * Terms can have spaces or any special character, it will properly
 * be detected.
 *
 * \bug
 * Dashes, spaces, and a few other characters, are considered
 * separators and some will not be properly handled.
 *
 * \bug
 * At this time this function does not automatically handle words in
 * plurial. If you have a word such as "world" in your taxonomy but
 * "worlds" in the node body, it will not be detected unless you have
 * worlds in your term synonyms.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in,out] $node The node to search and which terms are updated
 * \param[in] $terms The terms that are available in the vocabulary
 */
function _mo_auto_add_terms_to_node($vid, &$node, $terms) {
  // start with an empty haystack
  $haystack = '';

  // if user wants the title included, add it now
  $include_title = _mo_auto_add_terms_title($vid, 'test');
  if ($include_title) {
    $haystack .= ' ' . check_plain($node->title);
  }

  // if user wants the body included, add it now
  $include_body = _mo_auto_add_terms_body($vid, 'test');
  if ($include_body) {
    // XXX we probably want to do a node_view() and use the output instead
    //     at this time we get stuff like [toc] and [collapsed] included...
    if (_mo_auto_add_terms_render($vid, 'test')) {
      $body = node_view($node, FALSE, TRUE, FALSE);
    }
    else {
      $body = $node->body;
    }
    // make sure to completely remove scripts!
    // (the strip_tags() removes the start & end tags but not the script!)
    $body = preg_replace(
      array(
        '%<script.*</script\s*>%',
        '%<(div|p|h1|h2|h3|h4|h5|h6|br|pre|table|tr|th|td)%',
      ),
      array(
        '',
        ' <\1',
      ),
      $body);

    $haystack .= ' ' . strip_tags($body);
  }

  // should we include comments?
  if ($node->comment) {
    $include_comments = _mo_auto_add_terms_comments($vid, 'test');
    if ($include_comments) {
      // check only published comments for this node, but all of them...
      $sql = 'SELECT subject, comment FROM {comments} WHERE nid = %d AND status = %d';
      $result = db_query($sql, $node->nid, COMMENT_PUBLISHED);
      $comments = '';
      while ($comment = db_fetch_object($result)) {
        // generate a long string, include subject & body
        $comments .= $comment->subject . ' ' . $comment->comment . ' ';
      }
      $haystack .= ' ' . strip_tags($comments);
    }
  }

  // transform the haystack to something more usable (clean double spaces, etc.)
  $haystack = mo_autoaddterm_clean_term($haystack, $vid);
  if (!trim($haystack)) {
    // the haystack is empty! unlikely to happen but someone who
    // unselect the title, body & comments...
    return;
  }

  $has_term = FALSE;
  $matched = array();
  $other_matches = array();
  $vid_matches = array();
  $node_terms = array();
  $min_count = variable_get('mo_auto_add_terms_min_count' . $vid, 0);
  $max_count = variable_get('mo_auto_add_terms_max_count' . $vid, 0);
  $count_occurrences = $max_count > 0;
  $parents_only = _mo_auto_add_terms_only_parents($vid, 'test');
  $cond = !$parents_only && _mo_auto_add_terms_ifexist($vid, 'test') ? '' : ' AND v.vid <> %d';
  $check_composed_terms = _mo_auto_add_terms_check_composed_terms($vid, 'test');

  // load all the vocabularies (we do not want to mess up the others)
  //
  // Note: We directly handle the term_data table because the taxonomy
  //       functions do not handle them the way we need it to do so.
  //       Plus that way we can optimize the loading quite some.

  $sql = 'SELECT t.tid, t.vid FROM {term_node} r'
      . ' INNER JOIN {term_data} t ON r.tid = t.tid'
      . ' INNER JOIN {vocabulary} v ON t.vid = v.vid'
    . ' WHERE r.vid = %d ' . $cond;
  $result = db_query(db_rewrite_sql($sql, 't', 'tid'), $node->vid, $vid);
  while ($row = db_fetch_array($result)) {
    if ($row['vid'] == $vid) {
      $has_term = TRUE;
    }
    $matched[$row['tid']] = 1;
    $other_matches[] = $row['tid'];
  }

  // auto-add terms found in the node
  foreach ($terms as $tid => $synonyms) {
    if ($count_occurrences || $check_composed_terms) {
      $pattern = '';
      foreach ($synonyms as $needle) {
        $pattern .= '|' . preg_quote($needle, '/');
      }
      $pattern = '/' . substr($pattern, 1) . '/';
      if ($check_composed_terms) {
        $haystack = preg_replace($pattern, ' -- ', $haystack, -1, $count);
      }
      else {
        $count = preg_match_all($pattern, $haystack, $ignore);
      }
      if ($count > 0) {
        if (isset($matched[$tid])) {
          $matched[$tid] += $count;
        }
        else {
          $matched[$tid] = $count;
          $vid_matches[] = $tid;
        }
      }
    }
    elseif (empty($matched[$tid])) {
      foreach ($synonyms as $needle) {
        if (strpos($haystack, $needle) !== FALSE) {
          $has_term = TRUE;
          $matched[$tid] = 1;
          $vid_matches[] = $tid;
          break;
        }
      }
    }
  }

  if (!$has_term) {
    // Handle vocabularies that require at least one selection
    // Although that may fail, it is a good fallback solution in
    // most cases
    $vid_matches[] = _mo_auto_add_terms_auto_required($vid);
  }

  if ($count_occurrences) {
    // make sure we count only unique instances
    arsort($matched);
    $count = $max_count;
    $large_matches = array();
    $last_found = 0;
    foreach ($matched as $tid => $found) {
      if ($last_found != $found && $count <= 0) {
        break;
      }
      --$count;
      $large_matches[] = $tid;
      $last_found = $found;
    }
    $vid_matches = $large_matches;
  }

  // make sure we don't have any duplicates
  $vid_matches = array_unique($vid_matches);
  $matches = array();
  if ($parents_only) {
    // keep the parents only
    foreach ($vid_matches as $tid) {
      $parents = taxonomy_get_parents($tid);
      foreach ($parents as $p) {
        $matches[] = $p->tid;
      }
    }
  }
  elseif (_mo_auto_add_terms_parents($vid, 'test')) {
    // add parent terms for this vocabulary terms if any
    foreach ($vid_matches as $tid) {
      $parents = taxonomy_get_parents_all($tid);
      foreach ($parents as $p) {
        $matches[] = $p->tid;
      }
    }
  }
  else {
    // use terms as is
    $matches = $vid_matches;
  }

  $matches = array_unique($matches);
  $count = count($matches);
  if ($min_count > 0 && $count < $min_count) {
    // underflow, trigger event
    mo_auto_add_terms_do_actions('minimum_not_met', $node);
  }
  elseif ($max_count > 0 && $count > $max_count) {
    // overflow, trigger event
    mo_auto_add_terms_do_actions('maximum_not_met', $node);
  }

  $matches = array_unique(array_merge($other_matches, $matches));

  // save the result
  taxonomy_node_save($node, $matches);

  // reload in case another module needs to work on the terms as changed by us
  node_load($node->nid, NULL, TRUE);
}

/**
 * \brief Load all the terms from a vocabulary
 *
 * Load all the terms of a given vocabulary.
 *
 * The terms are ensured unique and if the same vocabulary is
 * loaded multiple times, the function returns the cached version.
 *
 * \param[in] $vid The vocabulary identifier
 *
 * \return An array with terms and their synonyms indexed by tid.
 */
function _mo_auto_add_terms_getallterms($vid) {
  static $s = array();

  // cached?
  if (isset($s[$vid])) {
    return $s[$vid];
  }

  // get all the terms of a taxonomy
  $term_object_array = taxonomy_get_tree($vid);

  $synonyms = array();
  foreach ($term_object_array as $term) {
    $tid = $term->tid;
    $term_list = array();

    // get the equivalents
    $result = db_query("SELECT mo_auto_add_terms_enabled, mo_auto_add_terms_equivalents"
                      . " FROM {mo_auto_add_terms_terms} WHERE tid = %d", $tid);
    if ($result) {
      $extras = db_fetch_object($result);
      if ($extras) {
        // is the term disabled? if so, skip it
        if (!$extras->mo_auto_add_terms_enabled) {
          continue;
        }
        $equivalents = preg_split('/[\r\n]+/', trim($extras->mo_auto_add_terms_equivalents));
        foreach ($equivalents as $e) {
          $term_list[] = mo_autoaddterm_clean_term($e, $vid);
        }
      }
    }

    // put the term itself
    $term_list[] = mo_autoaddterm_clean_term($term->name, $vid);

    if (_mo_auto_add_terms_synonyms($vid, 'test')) {
      // retrieve the corresponding synonyms
      $fetch_synonym_list = taxonomy_get_synonyms($tid);
      foreach ($fetch_synonym_list as $syn) {
        $term_list[] = mo_autoaddterm_clean_term($syn, $vid);
      }
    }

    // save the result, we keep only unique words (no need for duplicates)
    $synonyms[$tid] = array_unique($term_list);
  }

  // need sorting by length?
  if (_mo_auto_add_terms_check_composed_terms($vid, 'test')) {
    uasort($synonyms, '_mo_auto_add_terms_sort_by_length');
  }

  $s[$vid] = $synonyms;

  return $synonyms;
}

/**
 * \brief Sort two strings by length.
 *
 * This function is used to order two strings using their length,
 * the longest first.
 *
 * The idea is that composed terms must be searched first. That way
 * we can avoid counting a composed terms and each of its sub-term.
 *
 * \param[in] $a The left array of strings (before)
 * \param[in] $b The right array of strings (after)
 *
 * \return 0 when equal in length, -1 if $a is longer than $b and
 *   1 if $b is longer than $a.
 */
function _mo_auto_add_terms_sort_by_length($a, $b) {
  $la = 0;
  foreach ($a as $s) {
    $len = strlen($s);
    if ($len > $la) {
      $la = $len;
    }
  }
  $lb = 0;
  foreach ($b as $s) {
    $len = strlen($s);
    if ($len > $lb) {
      $lb = $len;
    }
  }
  if ($la < $lb) {
    return 1;
  }
  if ($la > $lb) {
    return -1;
  }
  return 0;
}

/**
 * \brief Reset the auto-addterms cache.
 *
 * Define a vocabulary and a node to delete the cache for that
 * vocabulary attached to that node.
 *
 * Define only the vocabulary when a given vocabulary was modified
 * in some way, this then deletes all the nodes that were processed
 * and that are linked to that vocabulary.
 *
 * Define only a node when it was modified to reset all the caches
 * whatever the vocabulary.
 *
 * Use null for both, $vid and $nid to delete the entire cache.
 *
 * \param[in] $vid The vocabulary that was modified
 * \param[in] $nid The specific node that was modified
 */
function _mo_auto_add_terms_reset_cache($vid, $nid) {
  $nid = (int)$nid; // we use $where below and must make sure $nid is a number
  if (is_null($vid)) {
    // delete all the nodes or all the node but $nid
    if ($nid) {
      $where_nid = " WHERE nid <> $nid";
    }
    else {
      $where_nid = '';
    }
    db_query("DELETE FROM {mo_auto_add_terms_nodes}"
            . $where_nid);
  }
  else {
    // delete from our cache all the nodes that have a type included in
    // the corresponding vocabulary
    if ($nid) {
      $and_nid = " AND nid <> $nid";
    }
    else {
      $and_nid = '';
    }
    db_query("DELETE FROM {mo_auto_add_terms_nodes} WHERE nid IN"
        . " (SELECT n.nid FROM {node} AS n, {vocabulary_node_types} AS t "
           . " WHERE t.vid = %d AND t.type = n.type)"
           . " $and_nid", $vid);
  }
}

/**
 * \brief Check whether the vocabulary is to be hidden in the node edit/add forms
 *
 * Check whether the user asked for that vocabulary to be hidden on the
 * node edit and add forms.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as hidden
 */
function _mo_auto_add_terms_hide($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_hide', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_hide', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_hide', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set, this way we return TRUE or FALSE
    $result = !empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_hide().', 'error');
    return FALSE;

  }

  return $result;
}

/**
 * \brief Check whether the vocabulary is to be used with auto-addterms
 *
 * Check whether the user selected that vocabulary as an auto-add-terms
 * vocabulary.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as auto-addterm aware
 */
function _mo_auto_add_terms_selected($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_selected', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_selected', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_selected', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set, this way we return TRUE or FALSE
    $result = !empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_selected().', 'error');
    return FALSE;

  }

  return $result;
}

/**
 * \brief Check whether the vocabulary is to be used with the auto-addterms CRON feature
 *
 * The user may want the auto add terms feature whenever he saves
 * nodes, but not with CRON. This is on a per vocabulary flag so
 * one vocabulary can be run with CRON and another can be turned
 * off.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as auto-addterm aware
 */
function _mo_auto_add_terms_autocheck($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_autocheck', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_autocheck', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_autocheck', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set meaning that we have to return FALSE
    $result = empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_autocheck().', 'error');
    return FALSE;

  }

  return $result;
}

/**
 * \brief Check whether the vocabulary should include the title for auto-add terms
 *
 * The module checks the title by default. It is possible for the user
 * to explicitly ask for the title to be included or to exclude the title.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as 'use title for auto add term'
 */
function _mo_auto_add_terms_title($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_title', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_title', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_title', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set meaning that we have to return FALSE
    $result = empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_title().', 'error');
    return FALSE;

  }

  return $result;
}


/**
 * \brief Check whether the vocabulary should include the body for auto-add terms
 *
 * The module uses the body by default. It is possible for the user
 * to ask for the body to be excluded. When that is done, the body
 * is not scanned for terms.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as 'use body for auto-add terms'
 */
function _mo_auto_add_terms_body($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_body', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_body', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_body', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set meaning that we have to return FALSE
    $result = empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_body().', 'error');
    return FALSE;

  }

  return $result;
}


/**
 * \brief Check whether the vocabulary should include comments for auto-add terms
 *
 * The module ignores the comments by default. It is possible for the user
 * to ask for the comments to be included. When that is done, the comments
 * are also scanned for terms. This can be a huge drag in regard to speed.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as 'remove unless it exists'
 */
function _mo_auto_add_terms_comments($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_comments', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_comments', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_comments', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set, this way we return TRUE or FALSE
    $result = !empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_comments().', 'error');
    return FALSE;

  }

  return $result;
}


/**
 * \brief Check whether the node should be rendered before processed.
 *
 * It is now possible to ask the module to render the node (i.e. calling
 * the node_view() function,) instead of using the body data as is. This
 * is important if you use many tags such as [view], [node], [block], etc.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary requests that we render the node
 */
function _mo_auto_add_terms_render($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_render', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_render', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_render', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set, this way we return TRUE or FALSE
    $result = !empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_render().', 'error');
    return FALSE;

  }

  return $result;
}


/**
 * \brief Check whether the vocabulary has auto add existing terms only
 *
 * By default, terms added by the users remain untouched and the module
 * adds terms present in both: the vocabulary and the node.
 *
 * Some users may want to eliminate any term that does not exist in the
 * node. For instance, if you add the term Yellow, write something about
 * Yellow and later delete that paragraph, then the term Yellow should
 * probably be removed from the list of terms. This feature does that just
 * that. However, at this time, it also prevents you from having terms
 * with spaces so it has to be an option.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as 'remove unless it exists'
 */
function _mo_auto_add_terms_ifexist($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_ifexist', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_ifexist', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_ifexist', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set, this way we return TRUE or FALSE
    $result = !empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_ifexist().', 'error');
    return FALSE;

  }

  return $result;
}


/**
 * \brief Check whether the vocabulary will make use of the synonyms
 *
 * By default, the MO Auto add terms ignores synonyms. If you want to
 * include synonyms as terms, select this flag.
 *
 * This is a new feature since version 1.7. Previous versions would
 * use synonyms automatically. Now the MO Auto add terms offers a
 * list of equivalents instead of synonyms. That way one can use
 * synonyms for other purposes such as displaying them on their
 * Taxonomy VTN index.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as 'include synonyms'
 */
function _mo_auto_add_terms_synonyms($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_synonyms', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_synonyms', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_synonyms', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set, this way we return TRUE or FALSE
    $result = !empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_synonyms().', 'error');
    return FALSE;

  }

  return $result;
}


/**
 * \brief Check whether the parents of a term are automatically added
 *
 * Normally, just a matched term is added. At times, it is useful to
 * also include all the parents of a term so someone can find that
 * node at all levels.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as 'add parents'
 */
function _mo_auto_add_terms_parents($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_parents', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_parents', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_parents', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set, this way we return TRUE or FALSE
    $result = !empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_parents().', 'error');
    return FALSE;

  }

  return $result;
}


/**
 * \brief Check whether only the parents matching terms are kept.
 *
 * When parents are searched, one can also just keep the parents. In
 * this case, the matching terms are discarded (unless one of them
 * is a parent of a matching term, then it is kept.)
 *
 * This is particularly useful when you have a one level parent/child
 * taxonomy to just tag pages with group like terms.
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary is marked as 'only parents'
 */
function _mo_auto_add_terms_only_parents($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_only_parents', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_only_parents', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_only_parents', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set, this way we return TRUE or FALSE
    $result = !empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_only_parents().', 'error');
    return FALSE;

  }

  return $result;
}

/**
 * \brief Check whether matching terms are removed as we count them.
 *
 * By default, the haystack is left untouched. When this flag is set
 * to true, the terms found in the haystack are also replaced by '--'.
 * This way, they do not match more than once. It helps for composed
 * words such as "Blue Car" which then won't match "Blue" nor "Car".
 *
 * \param[in] $vid The vocabulary identifier
 * \param[in] $func The action to be applied: 'set', 'clear', 'test'
 *
 * \return TRUE if that vocabulary terms shall be removed as found.
 */
function _mo_auto_add_terms_check_composed_terms($vid, $func) {
  static $mo_auto_add_terms_vid;

  if (!$mo_auto_add_terms_vid) {
    $mo_auto_add_terms_vid = variable_get('mo_auto_add_terms_check_composed_terms', array());
  }

  switch ($func) {
  case 'set':
    $mo_auto_add_terms_vid[$vid] = 1;
    variable_set('mo_auto_add_terms_check_composed_terms', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = TRUE;
    break;

  case 'clear':
    unset($mo_auto_add_terms_vid[$vid]);
    variable_set('mo_auto_add_terms_check_composed_terms', $mo_auto_add_terms_vid);
    _mo_auto_add_terms_reset_cache($vid, NULL);
    $result = FALSE;
    break;

  case 'test':
    // if equal to 1, it is set, this way we return TRUE or FALSE
    $result = !empty($mo_auto_add_terms_vid[$vid]);
    break;

  default:
    drupal_set_message('INTERNAL ERROR: invalid function ' . check_plain($func) . ' in _mo_auto_add_terms_check_composed_terms().', 'error');
    return FALSE;

  }

  return $result;
}

/**
 * \brief Check whether at least one term is required.
 *
 * Checks to see if required is checked on the specified vocabulary.
 * If so, returns the identifier of the term with the greatest weight.
 *
 * If the vocabulary requires a term but has none defined, then
 * the function fails by returning NULL.
 *
 * \param[in] $vid The vocabulary to be checked
 *
 * \return The $tid value (the term identifier)
 */
function _mo_auto_add_terms_auto_required($vid) {
  // cache result
  static $misc;

  if (is_null($misc)) {
    $result = db_query_range("SELECT a.tid FROM {term_data} AS a, {vocabulary} AS b"
       . " WHERE a.vid = b.vid AND b.required = 1 AND a.vid = %d"
       . " ORDER BY a.weight DESC", $vid, 1);
    if ($result) {
      $misc = db_fetch_array($result);
    }
    else {
      $misc['oops'] = 0;
    }
  }
  return $misc['tid'];
}

/**
 * \brief Process a few nodes.
 *
 * This function is called by the CRON system to process a few
 * nodes and attribute their terms as expected.
 *
 * \bug
 * We use a ring type of a test to check the nodes by group of
 * 10 or so nodes (10 by default, see the global settings to
 * change the default). If you are using PostgreSQL (and this
 * possibly happen with other database systems,) there are times
 * when you may only check one node at a time! We could add an
 * ORDER BY so as to avoid this problem, but that would require
 * a lot more processing time so instead we have this potential
 * bug. Increasing the number of nodes to be tested may help
 * since once in a while the NID will revert to zero and then
 * many more nodes can be worked on at once.
 *
 * \param[in] $is_cron TRUE if called by CRON, FALSE otherwise
 * \param[in] $count The number of nodes to work on, 0 to read the user defined number
 */
function _mo_auto_add_terms_auto_process_nodes($is_cron = FALSE, $count = 1000000) {
  if (!$count) {
    // Max. number of nodes to check on this round
    $count = variable_get('mo_auto_add_terms_node_count', AUTOADDTERMS_NODE_COUNT);
    if ($count <= 0) {
      // the user globally turned the CRON feature off
      return;
    }
  }
  elseif ($count >= 1000000) {
    // Try to allocate enough time to run all the hook_cron implementations.
    if (function_exists('set_time_limit')) {
      // let it run for up to 30 min.
      @set_time_limit(1800);
    }
  }
  if ($count < 5) {
    $count = 5;
  }

  // This is going to be a rather slow query and we have more coming below, so we limit it much
  $max_nid = variable_get('mo_auto_add_terms_max_nid', 0);
  $last_nid = $max_nid;

  $sql = "SELECT nid FROM {node} WHERE nid > %d AND nid NOT IN (SELECT nid FROM {mo_auto_add_terms_nodes})";
  $results = db_query_range($sql, $max_nid, 0, $count);

  // reset max, if no node were found, then we'll get 0
  // in the variable for next time!
  $max_nid = 0;
  $processed = 0;
  while ($row = db_fetch_array($results)) {
    if ($row['nid'] > $max_nid) {
      $max_nid = $row['nid'];
    }
    $node = node_load($row['nid']);
    if ($node) {
      _mo_auto_add_terms_process_node($node, FALSE, $is_cron);
      ++$processed;
      if ($processed % 10 == 0) {
        // just in case we take too much time...
        variable_set('mo_auto_add_terms_max_nid', $max_nid);
      }
    }
  }
  if ($max_nid < $last_nid) {
    // mark that all the nodes have been worked on
    variable_set('mo_auto_add_terms_looped', TRUE);
  }
  variable_set('mo_auto_add_terms_max_nid', $max_nid);

  if (!$is_cron) {
    if ($processed) {
      drupal_set_message($processed . ' nodes were processed. If you know you have more nodes, click the link again to process more nodes.', 'warning');
    }
    else {
      drupal_set_message('No nodes were processed. You are up to date.', 'warning');
    }
    drupal_goto('admin/settings/mo_auto_add_terms');
  }
}

/**
 * \brief Check whether a user has permission to see the terms while editing.
 *
 * Whenever a user edits or adds a node, he/she may not be allowed to
 * edit (add) terms to that node.
 *
 * The administrator can assign a role that will have the right to do so.
 *
 * \param[in] $vid The vocabulary being checked
 *
 * \return TRUE if the user has the right to edit (i.e. do not hide if true)
 */
function _mo_auto_add_terms_nohide_for_role($vid) {
  global $user;
  static $nohide_for_role = array();

  $name = 'mo_auto_add_terms_nohide_for_role' . $vid;
  if (!isset($nohide_for_role[$vid])) {
    $nohide_for_role[$vid] = variable_get($name, 0);
  }

  // no role is allowed to see the terms in the node/add/% & node/%/edit
  if ($nohide_for_role[$vid] == 0) {
    return FALSE;
  }

  return isset($user->roles[$nohide_for_role[$vid]]);
}

/**
 * \brief Hide taxonomies from the node form if required
 *
 * This function loops through the list of taxonomies and
 * remove those that the user asked to hide.
 *
 * When none of the taxonomies are marked as hidden, then
 * this function does nothing.
 *
 * \note
 * The loop looks complex because the D6 list is composed of
 * 'vid' and 'tags' with sub-array having their index set to
 * a 'vid'.
 *
 * \param[in,out] $form The form that will be modified
 */
function _mo_auto_add_terms_fix_node_form(&$form) {
  static $auto_hide = array();

  // any taxonomy set?
  if (!isset($form['taxonomy'])) {
    return;
  }

  $node_type = $form['type']['#value'];
  $hide_all = TRUE;
  foreach ($form['taxonomy'] as $vid => $value) {
    // skip direct fields
    if ($vid[0] == '#') {
      continue;
    }
    // In D6 tags are saved in a set of sub-arrays, just to annoy us
    if ($vid == 'tags') {
      foreach ($value as $vid => $ignore) {
        // user authorized to see these terms?
        if (_mo_auto_add_terms_nohide_for_role($vid)) {
          $hide_all = FALSE;
          continue;
        }
        if (!isset($auto_hide[$vid])) {
          $auto_hide[$vid] = _mo_auto_add_terms_hide($vid, 'test');
          if (!$auto_hide[$vid]) {
            $auto_hide[$vid] = _mo_auto_add_terms_content_type_hide($vid, $node_type);
          }
        }
        if ($auto_hide[$vid]) {
          unset($form['taxonomy']['tags'][$vid]);
        }
        else {
          $hide_all = FALSE;
        }
      }
    }
    else {
      // user authorized to see these terms?
      if (_mo_auto_add_terms_nohide_for_role($vid)) {
        $hide_all = FALSE;
      }
      else {
        if (!isset($auto_hide[$vid])) {
          $auto_hide[$vid] = _mo_auto_add_terms_hide($vid, 'test');
          if (!$auto_hide[$vid]) {
            $auto_hide[$vid] = _mo_auto_add_terms_content_type_hide($vid, $node_type);
          }
        }
        if ($auto_hide[$vid]) {
          unset($form['taxonomy'][$vid]);
        }
        else {
          $hide_all = FALSE;
        }
      }
    }
  }
  // if all the taxonomies were hidden and there was a fieldset, hide the fieldset
  if ($hide_all && isset($form['taxonomy']['#type'])) {
    $form['taxonomy']['#type'] = 'hidden';
  }
}

/**
 * \brief Check whether the taxonomy should be hidden.
 *
 * This function checks whether the user asked for this taxonomy to be
 * hidden according to the corresponding node type.
 *
 * \param[in] $vid The vocabulary identifier.
 * \param[in] $node_type The type of node that is being edited or added.
 *
 * \return Whether that vocabulary field should be hidden.
 */
function _mo_auto_add_terms_content_type_hide($vid, $node_type) {
  $hide = variable_get('mo_auto_add_terms_content_type_hide' . $vid, array());
  return !empty($hide[$node_type]);
}

/**
 * \brief Add a few fields in the taxonomy vocabulary form.
 *
 * This function adds settings in the taxonomy that the user can tweak
 * to change the behavior of the auto-add terms module.
 *
 * \param[in,out] $form The form that will be modified
 */
function _mo_auto_add_terms_edit_taxonomy_form_vocabulary(&$form) {
  // on "Add vocabulary" 'vid' is not yet defined
  $vid = isset($form['vid']['#value']) ? $form['vid']['#value'] : 0;

  $node_types = array_map('check_plain', node_get_types('names'));

  $content_type_hide = variable_get('mo_auto_add_terms_content_type_hide' . $vid, array());
  $form['content_types']['mo_auto_add_terms_content_type_hide'] = array(
    '#type' => 'checkboxes',
    '#title' => t('Types where vocabulary is hidden on node/add and node/%%/edit pages'),
    '#default_value' => $content_type_hide,
    '#options' => $node_types,
    '#description' => t('Instead of using the global Hide vocabulary flag (see below,) you can hide the taxonony on a per content type basis. To do so, do not check the global flag, instead, check the node types where you want the taxonomy box to be hidden.<br /><strong>IMPORTANT:</strong> Hiding a node type that is not selected for this taxonomy has no effect. Hiding a taxonomy is not the same as not selecting it. A taxonomy entry that is required should probably not be hidden.'),
  );

  $auto_hide = _mo_auto_add_terms_hide($vid, 'test');
  $form['content_types']['mo_auto_add_terms_hide'] = array(
    '#type' => 'checkbox',
    '#title' => t('Hide vocabulary on node/add and node/%/edit pages whatever the node type'),
    '#default_value' => $auto_hide,
    '#description' => t('The MO Auto add terms module also gives you the ability to hide the taxonomy fields from the node forms. This is useful if you do not want your users to add terms to this vocabulary.'),
  );

  // Note: no need for 'any' since we can have the user turn off the Hide flag
  $raw_roles = user_roles(TRUE, 'edit terms');  // we always ignore Anonymous users
  $roles = array(0 => 'none');
  foreach ($raw_roles as $k => $v) {
    $roles[$k] = $v;
  }
  $no_hide = variable_get('mo_auto_add_terms_nohide_for_role' . $vid, 0);
  $form['content_types']['mo_auto_add_terms_nohide_for_role'] = array(
    '#type' => 'select',
    '#title' => t('Do not hide the vocabulary if the user is part of this role'),
    '#default_value' => $no_hide,
    '#options' => $roles,
    '#description' => t('In order for some administrators or moderators to assign terms to a node that otherwise has its terms hidden, you can select the role that will be authorized to add terms to a node. When a user who is part of that role edits a node, the "Hide vocabulary on node/add and node/edit pages" will have no effect.')
       . '<br/>'
       . t('WARNING: Role must have the "edit terms" permission in order to be assigned this permission.'),
  );

  $form['mo_auto_add_terms'] = array(
    '#type' => 'fieldset',
    '#title' => t('MO Auto add terms'),
    '#collapsible' => TRUE,
    '#collapsed' => FALSE,
    '#description' => t('Setup the automatic addition of terms to your nodes.'),
  );

  $auto_selected = _mo_auto_add_terms_selected($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_selected'] = array(
    '#type' => 'checkbox',
    '#title' => t('Auto add terms <span style="color: red;">(REQUIRES MULTIPLE)</span>'),
    '#default_value' => $auto_selected,
    '#description' => t('This flag authorizes this module to automatically assign all the existing terms to a node when these are found in its title or body.'),
  );

  $form['mo_auto_add_terms']['mo_auto_add_terms_process_per_contenttype'] = array(
    '#type' => 'fieldset',
    '#title' => t('Limit processing by categories'),
    '#collapsible' => TRUE,
    '#collapsed' => TRUE,
  );

  $auto_add_terms_per_contenttype = variable_get('mo_auto_add_terms_auto_add_terms_per_contenttype' . $vid, array());
  foreach ($node_types as $type_name => $ignore) {
    if (!isset($auto_add_terms_per_contenttype[$type_name])) {
      // by default, select the content types
      $auto_add_terms_per_contenttype[$type_name] = $type_name;
    }
  }
  $form['mo_auto_add_terms']['mo_auto_add_terms_process_per_contenttype']['mo_auto_add_terms_auto_add_terms_per_contenttype'] = array(
    '#type' => 'checkboxes',
    '#title' => t('Process limited by content type'),
    '#default_value' => $auto_add_terms_per_contenttype,
    '#options' => $node_types,
    '#description' => t('By default, MO Auto add terms processes all the nodes attached to this taxonomy. However, there are cases when a content type should not be processed. Use this selector to remove some content type from the list of tests.'),
  );

  $auto_autocheck = _mo_auto_add_terms_autocheck($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_autocheck'] = array(
    '#type' => 'checkbox',
    '#title' => t('Automatically check this vocabulary with CRON'),
    '#default_value' => $auto_autocheck,
    '#description' => t('When this checkbox is selected, the CRON system will check nodes attached to this vocabulary for terms that were not yet assigned.'
      . ' To turn off CRON on all the taxonomies, use the !settings.'
      . ' You can also !run.', array(
        '!settings' => l('global settings', 'admin/settings/mo_auto_add_terms'),
        '!run' => l('run a check now', 'admin/settings/mo_auto_add_terms/run'))),
  );

  $auto_title = _mo_auto_add_terms_title($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_title'] = array(
    '#type' => 'checkbox',
    '#title' => t('Check the title for terms'),
    '#default_value' => $auto_title,
    '#description' => t('When this checkbox is selected, the title is checked for terms. If anyone can create a new node, then any user will be able to add terms to your site.'),
  );

  $auto_body = _mo_auto_add_terms_body($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_body'] = array(
    '#type' => 'checkbox',
    '#title' => t('Check the body for terms'),
    '#default_value' => $auto_body,
    '#description' => t('When this checkbox is selected, the body is checked for terms. If anyone can create a new node, then any user will be able to add terms to your site.'),
  );

  $auto_comments = _mo_auto_add_terms_comments($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_comments'] = array(
    '#type' => 'checkbox',
    '#title' => t('Check the comments for terms'),
    '#default_value' => $auto_comments,
    '#description' => t('When this checkbox is selected, all the comments are checked for terms. Note that can dramatically slow down the website if you have a large quantity of comments. If anyone can add a comment, your users will be able to <em>add terms</em> without your approval.'),
  );

  $render = _mo_auto_add_terms_render($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_render'] = array(
    '#type' => 'checkbox',
    '#title' => t('Render the node before parsing it'),
    '#default_value' => $render,
    '#description' => t('Use this flag if you use many tags such as [toc], [collapse], [node], [view], etc.')
      . '<br /><strong>' . t('WARNING:') . '</strong> '
          . t('This is slow and uses a lot more memory. Use with caution.'),
  );

  $auto_ifexist = _mo_auto_add_terms_ifexist($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_ifexist'] = array(
    '#type' => 'checkbox',
    '#title' => t('Keep all terms even if not found anywhere in your node'),
    '#default_value' => $auto_ifexist,
    '#description' => t('When the Auto add terms check box is selected, this flag determines whether only terms found in the nodes are kept (if selected, the module will keep all the terms you specified even if not present in the node.)'),
  );

  $auto_synonyms = _mo_auto_add_terms_synonyms($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_synonyms'] = array(
    '#type' => 'checkbox',
    '#title' => t('Use the synonyms of a term as aliases of that term'),
    '#default_value' => $auto_synonyms,
    '#description' => t('By default, MO Auto add terms ignores synonyms (since v1.7). If you had an older version or prefer to use the Synonyms text area, select this box.'),
  );

  $user_pattern = variable_get('mo_auto_add_terms_ignore_pattern' . $vid, '[ -&(-,./:-?[-_{-~]');
  $form['mo_auto_add_terms']['mo_auto_add_terms_ignore_pattern'] = array(
    '#type' => 'textfield',
    '#title' => t('Regular expression used to remove unwanted characters'),
    '#default_value' => $user_pattern,
    '#description' => t('Define a perl pattern used to remove characters from the text being treated. The default is @default which removes most punctuation except dashes and apostrophes. You can add a dash at the very beginning of the [...] to remove those and an apostrophe at the start or end to also remove apostrophes: @all.',
                          array('@default' => '[ -&(-,./:-?[-_{-~]',
                                '@all' => '[- -&(-,./:-?[-_{-~\']'))
                     . '<br />' . t('<strong>WARNING:</strong> This expression will not be escaped before use. Make sure you know what you are doing before editing this value.'),
  );

  $min_count = variable_get('mo_auto_add_terms_min_count' . $vid, 0);
  $form['mo_auto_add_terms']['mo_auto_add_terms_min_count'] = array(
    '#type' => 'textfield',
    '#title' => t('Minimum number of terms'),
    '#default_value' => $min_count,
    '#description' => t('The module triggers an event whenever a node does not include at least this many terms from this vocabulary. Set to 0 to avoid events.'),
  );

  $max_count = variable_get('mo_auto_add_terms_max_count' . $vid, 0);
  $form['mo_auto_add_terms']['mo_auto_add_terms_max_count'] = array(
    '#type' => 'textfield',
    '#title' => t('Maximum number of terms'),
    '#default_value' => $max_count,
    '#description' => t('When setting the maximum to another value than 0, the module counts occurrences of all the terms. Then it selects the terms with the highest score including ties until it has at least Minimum terms or it reaches the maximum number allowed. If the inclusion of ties generates more terms than this maximum value, then the module triggers an event. Note: maximum must be larger or equal to minimum.'),
  );

  $auto_parents = _mo_auto_add_terms_parents($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_parents'] = array(
    '#type' => 'checkbox',
    '#title' => t('Add all the parents'),
    '#default_value' => $auto_parents,
    '#description' => t('Select this option to automatically include matching terms and all their parents.'),
  );

  $auto_parents = _mo_auto_add_terms_only_parents($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_only_parents'] = array(
    '#type' => 'checkbox',
    '#title' => t('Keep the parents only'),
    '#default_value' => $auto_parents,
    '#description' => t('Only attach the parents terms to your nodes. The terms that matched and that are not also a parent are discarded from the result. Setting this flag implies "Add all the parents".'),
  );

  $check_composed_terms = _mo_auto_add_terms_check_composed_terms($vid, 'test');
  $form['mo_auto_add_terms']['mo_auto_add_terms_check_composed_terms'] = array(
    '#type' => 'checkbox',
    '#title' => t('Distinguish composed terms'),
    '#default_value' => $check_composed_terms,
    '#description' => t('When turned on this feature counts the terms and as it does so, it removes them from the list of words found in the title, body and comments (i.e. the haystack.) This can be slow for very large pages and use a large amount of memory. The function requires the terms to be sorted from the largest to the smallest so the longest term is removed first. Note also that the length of each synonym is not taken in account. The longest is used...'),
  );

  $form['mo_auto_add_terms_selected']['#element_validate'] =  array('_mo_auto_add_terms_validate');
  $form['#submit'][] = '_mo_auto_add_terms_form_vocabulary_submit';
}

/**
 * \brief Defines a form to setup the MO Auto add terms module.
 *
 * This function creates a form for the MO Auto add terms
 * module in order to allow for a few global variables to be
 * set.
 *
 * \return An array representing the form.
 */
function mo_auto_add_terms_settings() {
  $form['mo_auto_add_terms']['cron'] = array(
    '#type' => 'fieldset',
    '#title' => t('CRON settings'),
    '#collapsible' => TRUE,
    '#collapsed' => FALSE,
    '#description' => t('This page lets you define the global settings of the MO Auto-add terms. For detailed settings, go to your !taxonomy instead.<br/>'
      . '<br/>'
      . '!run<br/>'
      . ' This check searches for up to 1 million of your nodes.'
      . ' If you have more nodes, click the link multiple times as required.'
      . ' Note that a check against all your nodes has the potential to time out.'
      . ' All you have to do is click the link again if you run in that problem.'
      . ' In general, a time out will result in a blank screen and no more activity.'
      . ' This check will run against all your taxonomies, even those without the <em>Automatically check this vocabulary with CRON</em> flag.',
        array(
          '!taxonomy' => l('taxonomies', 'admin/content/taxonomy'),
          '!run' => l('Run a check of all your nodes now', 'admin/settings/mo_auto_add_terms/run'),
        )),
  );

  $count = variable_get('mo_auto_add_terms_node_count', AUTOADDTERMS_NODE_COUNT);
  $form['mo_auto_add_terms']['cron']['mo_auto_add_terms_node_count'] = array(
    '#type' => 'textfield',
    '#title' => t('Nodes to process per CRON run'),
    '#description' => t('Defines the number of nodes you want the CRON process to parse on each run. If you have large nodes (a lot of text) think about using a smaller number. If you have small nodes and a large quantity of them, try to use a larger number. The default is 10. Use 0 to turn off the CRON capability altogether.'),
    '#maxlength' => 5,
    '#size' => 5,
    '#default_value' => $count,
  );

  $form['submit'] = array(
    '#type' => 'submit',
    '#value' => t('Save'),
    '#submit' => array('mo_auto_add_terms_form_settings_submit'),
  );

  return $form;
}

/**
 * \brief Return the form while editing terms.
 *
 * This function adds a few boxes to the Terms form in order for
 * the user to be able to define how to handle the term with the
 * MO Auto-Add Terms filter.
 *
 * \param[in,out] $form The form we want to alter
 */
function mo_auto_add_terms_terms_form(&$form) {
  $mo_auto_add_terms_enabled     = 1;
  $mo_auto_add_terms_equivalents = '';

  if (isset($form['tid']['#value'])) {
    // editing, get current data from database
    $tid = $form['tid']['#value'];

    $sql = "SELECT mo_auto_add_terms_enabled, mo_auto_add_terms_equivalents FROM {mo_auto_add_terms_terms} WHERE tid = %d";
    $result = db_query($sql, $tid);
    if ($row = db_fetch_array($result)) {
      $mo_auto_add_terms_enabled     = $row['mo_auto_add_terms_enabled'];
      $mo_auto_add_terms_equivalents = $row['mo_auto_add_terms_equivalents'];
    }
  }
  // else we're adding

  $form['identification']['mo_auto_add_terms_enabled'] = array(
    '#type' => 'checkbox',
    '#title' => t('Enabled in MO Auto add terms'),
    '#description' => t('If checked, this term will be taken in account by MO Auto add terms. By default all terms are enabled.'),
    '#default_value' => $mo_auto_add_terms_enabled,
  );

  $form['advanced']['mo_auto_add_terms_equivalents'] = array(
    '#type' => 'textarea',
    '#title' => t('MO Auto add terms Equivalents'),
    '#description' => t('Defines a list of equivalents (synonyms, plurial form, conjugated forms, misspelling, etc.)<br />Enter one equivalent per line.'),
    '#default_value' => $mo_auto_add_terms_equivalents,
    '#wysiwyg' => FALSE,
  );

  $form['#submit'][] = '_mo_auto_add_terms_terms_form_submit';
}

/**
 * \brief Call the MO Auto add terms actions.
 *
 * This function is used as the trigger that calls the user
 * defined MO Auto add terms actions.
 *
 * The available operations are as follow:
 *
 * \li minimum_not_met -- the minimum number of terms to attach to a node was not met
 * \li maximum_not_met -- too many terms need to be attached to a node (more than max. count)
 *
 * \param[in] $op The operation
 * \param[in] $node The problematic node
 */
function mo_auto_add_terms_do_actions($op, $node) {
  global $user;

  if (module_exists('trigger')) {
    $aids = _trigger_get_hook_aids('mo_auto_add_terms', $op);
    if (count($aids) > 0) {
      $author = user_load(array('uid' => $node->uid));
      if (!$author) {
        // in case it was deleted and not correctly defined in the node
        $author = $user;
      }
      $context = array(
        'hook' => 'mo_auto_add_terms',
        'op' => $op,
        'node' => $node,
        'author' => $author,
        'author_email' => $author->mail,
      );
      // with CRON user would be the anonymous user...
      if ($user->uid == 0) {
        $context['poster'] = $author;
        $context['poster_email'] = $author->mail;
      }
      else {
        $context['poster'] = $user;
        $context['poster_email'] = $user->mail;
      }
      foreach ($aids as $aid => $action_info) {
        actions_do($aid, $node, $context);
      }
    }
  }
}

/**
 * \brief Node type form additions.
 *
 * This function intercepts the vocabulary form and adds a few
 * widgets to handle the MO Auto add terms functionality.
 *
 * \param[in,out] $form The form to alter
 * \param[in] $form_state what state the form is in (i.e. submitted?)
 */
function mo_auto_add_terms_node_type_form(&$form, $form_state) {
  $settings = variable_get('mo_auto_add_terms_settings_node_' . $form['#node_type']->type, array());

  $form['mo_auto_add_terms'] = array(
    '#type' => 'fieldset',
    '#title' => t('MO Auto-add terms limits'),
    '#collapsible' => TRUE,
    '#collapsed' => TRUE,
  );
  $form['mo_auto_add_terms']['mo_auto_add_terms_title_error'] = array(
    '#type' => 'checkbox',
    '#title' => t('Title error'),
    '#description' => t('If selected, generate an error when the title is too short or too long. Otherwise generate a warning.'),
    '#default_value' => isset($settings['title_error']) ? $settings['title_error'] : 0,
  );
  $form['mo_auto_add_terms']['mo_auto_add_terms_title_minimum'] = array(
    '#type' => 'textfield',
    '#title' => t('Title minimum size'),
    '#description' => t('Enter the minimum size of the title required for a page of this type. If smaller, a warning or an error is generated when saving the page. Use 0 to ignore the minimum.'),
    '#default_value' => isset($settings['title_minimum']) ? $settings['title_minimum'] : 0,
  );
  $form['mo_auto_add_terms']['mo_auto_add_terms_title_maximum'] = array(
    '#type' => 'textfield',
    '#title' => t('Title maximum size'),
    '#description' => t('Enter the maximum size of the title required for a page of this type. If larger, a warning or an error is generated when saving the page. Use 0 to ignore the maximum. Otherwise make sure that the maximum is larger than the minimum.'),
    '#default_value' => isset($settings['title_maximum']) ? $settings['title_maximum'] : 0,
  );

  $form['#validate'][] = 'mo_auto_add_terms_node_settings_validate';
  $form['#submit'][] = 'mo_auto_add_terms_node_settings_submit';
}

/**
 * \brief Validate the node type form parameters.
 *
 * This function makes sure that minimum < maximum unless maximum is 0.
 * Also it makes sure that minimum and maximum are between 0 and 255 inclusive.
 *
 * \note
 * The maximum size of 255 is wrong if you consider UTF-8 since 255 UTF-8
 * characters will not fit in the database.
 *
 * \param[in] $form The form to validate.
 * \param[in] $form_state The current state of the form to validate.
 *
 * \sa mo_auto_add_terms_node_type_form
 */
function mo_auto_add_terms_node_settings_validate($form, &$form_state) {
  $values = &$form_state['values'];
  if ($values['mo_auto_add_terms_title_minimum'] > 0 && $values['mo_auto_add_terms_title_maximum'] > 0 && $values['mo_auto_add_terms_title_minimum'] > $values['mo_auto_add_terms_title_maximum']) {
    form_error($form['mo_auto_add_terms']['mo_auto_add_terms_title_maximum'], t('The maximum title size must be larger than the minimum size.'));
  }
  if ($values['mo_auto_add_terms_title_minimum'] < 0 || $values['mo_auto_add_terms_title_minimum'] > 255) {
    form_error($form['mo_auto_add_terms']['mo_auto_add_terms_title_minimum'], t('The minimum title size must be between 0 and 255.'));
  }
  if ($values['mo_auto_add_terms_title_maximum'] < 0 || $values['mo_auto_add_terms_title_maximum'] > 255) {
    form_error($form['mo_auto_add_terms']['mo_auto_add_terms_title_maximum'], t('The maximum title size must be between 0 and 255.'));
  }
}

/**
 * \brief Save the node type form parameters.
 *
 * This function saves the node type form parameters.
 *
 * \param[in] $form The form being submitted.
 * \param[in,out] $form_state The current state of the form including the user entered data.
 *
 * \sa mo_auto_add_terms_node_type_form
 */
function mo_auto_add_terms_node_settings_submit($form, &$form_state) {
  $values = &$form_state['values'];

  $settings = array(
    'title_error' => $values['mo_auto_add_terms_title_error'],
    'title_minimum' => (int)$values['mo_auto_add_terms_title_minimum'],
    'title_maximum' => (int)$values['mo_auto_add_terms_title_maximum'],
  );

  variable_set('mo_auto_add_terms_settings_node_' . $form['#node_type']->type, $settings);
}

/**
 * \brief Check the size of the node title.
 *
 * This function validates the size of the node title against the node type
 * minimum and maximum definitions.
 *
 * \param[in] $node The node.
 */
function _mo_auto_add_terms_validate_node($node) {
  $settings = variable_get('mo_auto_add_terms_settings_node_' . $node->type, array());

  // if the user specifies a title limit we expect the title to be defined...
  // but we still have to test whether it is there or not
  if (isset($settings['title_minimum'])) {
    $len = drupal_strlen($node->title);
    if ($len < $settings['title_minimum']) {
      if ($settings['title_error']) {
        form_set_error('title', 'Your title is too short. The minimum length is ' . $settings['title_minimum'] . '. The current length is ' . $len);
      }
      else {
        drupal_set_message('Your title is too short. The minimum length is expected to be ' . $settings['title_minimum'] . '. The current length is ' . $len, 'warning');
      }
    }
    elseif ($len > $settings['title_maximum'] && $settings['title_maximum'] > 0) {
      if ($settings['title_error']) {
        form_set_error('title', 'Your title is too long. The maximum length is ' . $settings['title_maximum'] . '. The current length is ' . $len);
      }
      else {
        drupal_set_message('Your title is too long. The maximum length is expected to be ' . $settings['title_maximum'] . '. The current length is ' . $len, 'warning');
      }
    }
  }
}

// vim: ts=2 sw=2 et syntax=php
